# pylint: skip-file
import threading
from unittest import mock

import ophyd
import pytest
from bec_lib import MessageEndpoints, messages

from ophyd_devices.epics.devices.eiger9m_csaxs import Eiger9McSAXS
from tests.utils import DMMock, MockPV


def patch_dual_pvs(device):
    for walk in device.walk_signals():
        if not hasattr(walk.item, "_read_pv"):
            continue
        if not hasattr(walk.item, "_write_pv"):
            continue
        if walk.item._read_pv.pvname.endswith("_RBV"):
            walk.item._read_pv = walk.item._write_pv


@pytest.fixture(scope="function")
def mock_det():
    name = "eiger"
    prefix = "X12SA-ES-EIGER9M:"
    sim_mode = False
    dm = DMMock()
    with mock.patch.object(dm, "connector"):
        with (
            mock.patch("ophyd_devices.epics.devices.psi_detector_base.FileWriter"),
            mock.patch(
                "ophyd_devices.epics.devices.psi_detector_base.PSIDetectorBase._update_service_config"
            ),
        ):
            with mock.patch.object(ophyd, "cl") as mock_cl:
                mock_cl.get_pv = MockPV
                mock_cl.thread_class = threading.Thread
                with mock.patch.object(Eiger9McSAXS, "_init"):
                    det = Eiger9McSAXS(
                        name=name, prefix=prefix, device_manager=dm, sim_mode=sim_mode
                    )
                    patch_dual_pvs(det)
                    yield det


def test_init():
    """Test the _init function:"""
    name = "eiger"
    prefix = "X12SA-ES-EIGER9M:"
    sim_mode = False
    dm = DMMock()
    with mock.patch.object(dm, "connector"):
        with (
            mock.patch("ophyd_devices.epics.devices.psi_detector_base.FileWriter"),
            mock.patch(
                "ophyd_devices.epics.devices.psi_detector_base.PSIDetectorBase._update_service_config"
            ),
        ):
            with mock.patch.object(ophyd, "cl") as mock_cl:
                mock_cl.get_pv = MockPV
                with (
                    mock.patch(
                        "ophyd_devices.epics.devices.eiger9m_csaxs.Eiger9MSetup.initialize_default_parameter"
                    ) as mock_default,
                    mock.patch(
                        "ophyd_devices.epics.devices.eiger9m_csaxs.Eiger9MSetup.initialize_detector"
                    ) as mock_init_det,
                    mock.patch(
                        "ophyd_devices.epics.devices.eiger9m_csaxs.Eiger9MSetup.initialize_detector_backend"
                    ) as mock_init_backend,
                ):
                    Eiger9McSAXS(name=name, prefix=prefix, device_manager=dm, sim_mode=sim_mode)
                    mock_default.assert_called_once()
                    mock_init_det.assert_called_once()
                    mock_init_backend.assert_called_once()


@pytest.mark.parametrize(
    "trigger_source, detector_state, expected_exception", [(2, 1, True), (2, 0, False)]
)
def test_initialize_detector(mock_det, trigger_source, detector_state, expected_exception):
    """Test the _init function:

    This includes testing the functions:
    - _init_detector
    - _stop_det
    - _set_trigger
    --> Testing the filewriter is done in test_init_filewriter

    Validation upon setting the correct PVs

    """
    mock_det.cam.detector_state._read_pv.mock_data = detector_state
    if expected_exception:
        with pytest.raises(Exception):
            mock_det.timeout = 0.1
            mock_det.custom_prepare.initialize_detector()
    else:
        mock_det.custom_prepare.initialize_detector()  # call the method you want to test
        assert mock_det.cam.acquire.get() == 0
        assert mock_det.cam.detector_state.get() == detector_state
        assert mock_det.cam.trigger_mode.get() == trigger_source


def test_trigger(mock_det):
    """Test the trigger function:
    Validate that trigger calls the custom_prepare.on_trigger() function
    """
    with mock.patch.object(mock_det.custom_prepare, "on_trigger") as mock_on_trigger:
        mock_det.trigger()
        mock_on_trigger.assert_called_once()


@pytest.mark.parametrize(
    "readout_time, expected_value", [(1e-3, 3e-3), (3e-3, 3e-3), (5e-3, 5e-3), (None, 3e-3)]
)
def test_update_readout_time(mock_det, readout_time, expected_value):
    if readout_time is None:
        mock_det.custom_prepare.update_readout_time()
        assert mock_det.readout_time == expected_value
    else:
        mock_det.scaninfo.readout_time = readout_time
        mock_det.custom_prepare.update_readout_time()
        assert mock_det.readout_time == expected_value


@pytest.mark.parametrize(
    "eacc, exp_url, daq_status, daq_cfg, expected_exception",
    [
        ("e12345", "http://xbl-daq-29:5000", {"state": "READY"}, {"writer_user_id": 12543}, False),
        ("e12345", "http://xbl-daq-29:5000", {"state": "READY"}, {"writer_user_id": 15421}, False),
        ("e12345", "http://xbl-daq-29:5000", {"state": "BUSY"}, {"writer_user_id": 15421}, True),
        ("e12345", "http://xbl-daq-29:5000", {"state": "READY"}, {"writer_ud": 12345}, True),
    ],
)
def test_initialize_detector_backend(
    mock_det, eacc, exp_url, daq_status, daq_cfg, expected_exception
):
    """Test self.custom_prepare.initialize_detector_backend (std daq in this case)

    This includes testing the functions:

    - _update_service_config

    Validation upon checking set values in mocked std_daq instance
    """
    with mock.patch("ophyd_devices.epics.devices.eiger9m_csaxs.StdDaqClient") as mock_std_daq:
        instance = mock_std_daq.return_value
        instance.stop_writer.return_value = None
        instance.get_status.return_value = daq_status
        instance.get_config.return_value = daq_cfg
        mock_det.scaninfo.username = eacc
        # scaninfo.username.return_value = eacc
        if expected_exception:
            with pytest.raises(Exception):
                mock_det.timeout = 0.1
                mock_det.custom_prepare.initialize_detector_backend()
        else:
            mock_det.custom_prepare.initialize_detector_backend()

            instance.stop_writer.assert_called_once()
            instance.get_status.assert_called()
            instance.set_config.assert_called_once_with(daq_cfg)


@pytest.mark.parametrize(
    "scaninfo, daq_status, daq_cfg, detector_state, stopped, expected_exception",
    [
        (
            {
                "eacc": "e12345",
                "num_points": 500,
                "frames_per_trigger": 1,
                "filepath": "test.h5",
                "scan_id": "123",
                "mokev": 12.4,
            },
            {"state": "READY"},
            {"writer_user_id": 12543},
            5,
            False,
            False,
        ),
        (
            {
                "eacc": "e12345",
                "num_points": 500,
                "frames_per_trigger": 1,
                "filepath": "test.h5",
                "scan_id": "123",
                "mokev": 12.4,
            },
            {"state": "BUSY"},
            {"writer_user_id": 15421},
            5,
            False,
            False,
        ),
        (
            {
                "eacc": "e12345",
                "num_points": 500,
                "frames_per_trigger": 1,
                "filepath": "test.h5",
                "scan_id": "123",
                "mokev": 18.4,
            },
            {"state": "READY"},
            {"writer_user_id": 12345},
            4,
            False,
            True,
        ),
    ],
)
def test_stage(
    mock_det, scaninfo, daq_status, daq_cfg, detector_state, stopped, expected_exception
):
    with (
        mock.patch.object(mock_det.custom_prepare, "std_client") as mock_std_daq,
        mock.patch.object(
            mock_det.custom_prepare, "publish_file_location"
        ) as mock_publish_file_location,
    ):
        mock_std_daq.stop_writer.return_value = None
        mock_std_daq.get_status.return_value = daq_status
        mock_std_daq.get_config.return_value = daq_cfg
        mock_det.scaninfo.num_points = scaninfo["num_points"]
        mock_det.scaninfo.frames_per_trigger = scaninfo["frames_per_trigger"]
        mock_det.filewriter.compile_full_filename.return_value = scaninfo["filepath"]
        # TODO consider putting energy as variable in scaninfo
        mock_det.device_manager.add_device("mokev", value=12.4)
        mock_det.cam.beam_energy.put(scaninfo["mokev"])
        mock_det.stopped = stopped
        mock_det.cam.detector_state._read_pv.mock_data = detector_state
        with mock.patch.object(mock_det.custom_prepare, "prepare_detector_backend") as mock_prep_fw:
            mock_det.filepath = scaninfo["filepath"]
            if expected_exception:
                with pytest.raises(Exception):
                    mock_det.timeout = 0.1
                    mock_det.stage()
            else:
                mock_det.stage()
                mock_prep_fw.assert_called_once()
                # Check _prep_det
                assert mock_det.cam.num_images.get() == int(
                    scaninfo["num_points"] * scaninfo["frames_per_trigger"]
                )
                assert mock_det.cam.num_frames.get() == 1

                mock_publish_file_location.assert_called_with(done=False)
                assert mock_det.cam.acquire.get() == 1


@pytest.mark.parametrize(
    "scaninfo, daq_status, expected_exception",
    [
        (
            {
                "eacc": "e12345",
                "num_points": 500,
                "frames_per_trigger": 1,
                "filepath": "test.h5",
                "scan_id": "123",
            },
            {"state": "BUSY", "acquisition": {"state": "WAITING_IMAGES"}},
            False,
        ),
        (
            {
                "eacc": "e12345",
                "num_points": 500,
                "frames_per_trigger": 1,
                "filepath": "test.h5",
                "scan_id": "123",
            },
            {"state": "BUSY", "acquisition": {"state": "WAITING_IMAGES"}},
            False,
        ),
        (
            {
                "eacc": "e12345",
                "num_points": 500,
                "frames_per_trigger": 1,
                "filepath": "test.h5",
                "scan_id": "123",
            },
            {"state": "BUSY", "acquisition": {"state": "ERROR"}},
            True,
        ),
    ],
)
def test_prepare_detector_backend(mock_det, scaninfo, daq_status, expected_exception):
    with (
        mock.patch.object(mock_det.custom_prepare, "std_client") as mock_std_daq,
        mock.patch.object(mock_det.custom_prepare, "filepath_exists") as mock_file_path_exists,
        mock.patch.object(mock_det.custom_prepare, "stop_detector_backend") as mock_stop_backend,
        mock.patch.object(mock_det, "scaninfo"),
    ):
        mock_std_daq.start_writer_async.return_value = None
        mock_std_daq.get_status.return_value = daq_status
        mock_det.filewriter.compile_full_filename.return_value = scaninfo["filepath"]
        mock_det.scaninfo.num_points = scaninfo["num_points"]
        mock_det.scaninfo.frames_per_trigger = scaninfo["frames_per_trigger"]

        if expected_exception:
            with pytest.raises(Exception):
                mock_det.timeout = 0.1
                mock_det.custom_prepare.prepare_data_backend()
                mock_file_path_exists.assert_called_once()
                assert mock_stop_backend.call_count == 2

        else:
            mock_det.custom_prepare.prepare_data_backend()
            mock_file_path_exists.assert_called_once()
            mock_stop_backend.assert_called_once()

        daq_writer_call = {
            "output_file": scaninfo["filepath"],
            "n_images": int(scaninfo["num_points"] * scaninfo["frames_per_trigger"]),
        }
        mock_std_daq.start_writer_async.assert_called_with(daq_writer_call)


@pytest.mark.parametrize("stopped, expected_exception", [(False, False), (True, True)])
def test_unstage(mock_det, stopped, expected_exception):
    with (
        mock.patch.object(mock_det.custom_prepare, "finished") as mock_finished,
        mock.patch.object(
            mock_det.custom_prepare, "publish_file_location"
        ) as mock_publish_file_location,
    ):
        mock_det.stopped = stopped
        if expected_exception:
            mock_det.unstage()
            assert mock_det.stopped is True
        else:
            mock_det.unstage()
            mock_finished.assert_called_once()
            mock_publish_file_location.assert_called_with(done=True, successful=True)
            assert mock_det.stopped is False


def test_stop_detector_backend(mock_det):
    with mock.patch.object(mock_det.custom_prepare, "std_client") as mock_std_daq:
        mock_std_daq.stop_writer.return_value = None
        mock_det.std_client = mock_std_daq
        mock_det.custom_prepare.stop_detector_backend()
        mock_std_daq.stop_writer.assert_called_once()


@pytest.mark.parametrize(
    "scaninfo",
    [
        ({"filepath": "test.h5", "successful": True, "done": False, "scan_id": "123"}),
        ({"filepath": "test.h5", "successful": False, "done": True, "scan_id": "123"}),
        ({"filepath": "test.h5", "successful": None, "done": True, "scan_id": "123"}),
    ],
)
def test_publish_file_location(mock_det, scaninfo):
    mock_det.scaninfo.scan_id = scaninfo["scan_id"]
    mock_det.filepath = scaninfo["filepath"]
    mock_det.custom_prepare.publish_file_location(
        done=scaninfo["done"], successful=scaninfo["successful"]
    )
    if scaninfo["successful"] is None:
        msg = messages.FileMessage(file_path=scaninfo["filepath"], done=scaninfo["done"])
    else:
        msg = messages.FileMessage(
            file_path=scaninfo["filepath"], done=scaninfo["done"], successful=scaninfo["successful"]
        )
    expected_calls = [
        mock.call(
            MessageEndpoints.public_file(scaninfo["scan_id"], mock_det.name),
            msg,
            pipe=mock_det.connector.pipeline.return_value,
        ),
        mock.call(
            MessageEndpoints.file_event(mock_det.name),
            msg,
            pipe=mock_det.connector.pipeline.return_value,
        ),
    ]
    assert mock_det.connector.set_and_publish.call_args_list == expected_calls


def test_stop(mock_det):
    with (
        mock.patch.object(mock_det.custom_prepare, "stop_detector") as mock_stop_det,
        mock.patch.object(
            mock_det.custom_prepare, "stop_detector_backend"
        ) as mock_stop_detector_backend,
    ):
        mock_det.stop()
        mock_stop_det.assert_called_once()
        mock_stop_detector_backend.assert_called_once()
        assert mock_det.stopped is True


@pytest.mark.parametrize(
    "stopped, scaninfo, cam_state, daq_status, expected_exception",
    [
        (
            False,
            {"num_points": 500, "frames_per_trigger": 4},
            0,
            {"acquisition": {"state": "FINISHED", "stats": {"n_write_completed": 2000}}},
            False,
        ),
        (
            False,
            {"num_points": 500, "frames_per_trigger": 4},
            0,
            {"acquisition": {"state": "FINISHED", "stats": {"n_write_completed": 1999}}},
            True,
        ),
        (
            False,
            {"num_points": 500, "frames_per_trigger": 1},
            1,
            {"acquisition": {"state": "READY", "stats": {"n_write_completed": 500}}},
            True,
        ),
        (
            False,
            {"num_points": 500, "frames_per_trigger": 1},
            0,
            {"acquisition": {"state": "FINISHED", "stats": {"n_write_completed": 500}}},
            False,
        ),
    ],
)
def test_finished(mock_det, stopped, cam_state, daq_status, scaninfo, expected_exception):
    with (
        mock.patch.object(mock_det.custom_prepare, "std_client") as mock_std_daq,
        mock.patch.object(mock_det.custom_prepare, "stop_detector_backend") as mock_stop_backend,
        mock.patch.object(mock_det.custom_prepare, "stop_detector") as mock_stop_det,
    ):
        mock_std_daq.get_status.return_value = daq_status
        mock_det.cam.acquire._read_pv.mock_state = cam_state
        mock_det.scaninfo.num_points = scaninfo["num_points"]
        mock_det.scaninfo.frames_per_trigger = scaninfo["frames_per_trigger"]
        if expected_exception:
            with pytest.raises(Exception):
                mock_det.timeout = 0.1
                mock_det.custom_prepare.finished()
                assert mock_det.stopped is stopped
        else:
            mock_det.custom_prepare.finished()
            if stopped:
                assert mock_det.stopped is stopped

            mock_stop_backend.assert_called()
            mock_stop_det.assert_called_once()



""" This module contains tests for the simulation devices in ophyd_devices """

# pylint: disable: all
import os
from unittest import mock

import h5py
import numpy as np
import pytest
from ophyd import Device, Signal

from ophyd_devices.ophyd_base_devices.bec_protocols import (
    BECDeviceProtocol,
    BECFlyerProtocol,
    BECPositionerProtocol,
    BECScanProtocol,
    BECSignalProtocol,
)
from ophyd_devices.sim.sim import SimCamera, SimFlyer, SimMonitor, SimPositioner
from ophyd_devices.sim.sim_frameworks import H5ImageReplayProxy, SlitProxy
from ophyd_devices.sim.sim_signals import ReadOnlySignal
from ophyd_devices.utils.bec_device_base import BECDevice, BECDeviceBase
from tests.utils import DMMock


@pytest.fixture(scope="function")
def signal(name="signal"):
    """Fixture for Signal."""
    sig = ReadOnlySignal(name=name, value=0)
    yield sig


@pytest.fixture(scope="function")
def monitor(name="monitor"):
    """Fixture for SimMonitor."""
    dm = DMMock()
    mon = SimMonitor(name=name, device_manager=dm)
    yield mon


@pytest.fixture(scope="function")
def camera(name="camera"):
    """Fixture for SimCamera."""
    dm = DMMock()
    cam = SimCamera(name=name, device_manager=dm)
    yield cam


@pytest.fixture(scope="function")
def positioner(name="positioner"):
    """Fixture for SimPositioner."""
    dm = DMMock()
    pos = SimPositioner(name=name, device_manager=dm)
    yield pos


@pytest.fixture(scope="function")
def h5proxy_fixture(name="h5proxy"):
    """Fixture for SimCamera."""
    dm = DMMock()
    proxy = H5ImageReplayProxy(name=name, device_manager=dm)
    camera = SimCamera(name="eiger", device_manager=dm)
    yield proxy, camera


@pytest.fixture(scope="function")
def slitproxy_fixture(name="slit_proxy"):
    """Fixture for SimCamera."""
    dm = DMMock()
    proxy = SlitProxy(name=name, device_manager=dm)
    camera = SimCamera(name="eiger", device_manager=dm)
    samx = SimPositioner(name="samx", device_manager=dm)
    yield proxy, camera, samx


@pytest.fixture(scope="function")
def flyer(name="flyer"):
    """Fixture for SimFlyer."""
    dm = DMMock()
    fly = SimFlyer(name=name, device_manager=dm)
    yield fly


def test_signal__init__(signal):
    """Test the BECProtocol class"""
    assert isinstance(signal, BECDeviceProtocol)
    assert isinstance(signal, BECSignalProtocol)


def test_monitor__init__(monitor):
    """Test the __init__ method of SimMonitor."""
    assert isinstance(monitor, SimMonitor)
    assert isinstance(monitor, BECDeviceProtocol)
    assert isinstance(monitor, BECScanProtocol)


def test_camera__init__(camera):
    """Test the __init__ method of SimMonitor."""
    assert isinstance(camera, SimCamera)
    assert isinstance(camera, BECDeviceProtocol)
    assert isinstance(camera, BECScanProtocol)


def test_positioner__init__(positioner):
    """Test the __init__ method of SimPositioner."""
    assert isinstance(positioner, SimPositioner)
    assert isinstance(positioner, BECDeviceProtocol)
    assert isinstance(positioner, BECScanProtocol)
    assert isinstance(positioner, BECPositionerProtocol)


def test_flyer__init__(flyer):
    """Test the __init__ method of SimFlyer."""
    assert isinstance(flyer, SimFlyer)
    assert isinstance(flyer, BECDeviceProtocol)
    assert isinstance(flyer, BECScanProtocol)
    assert isinstance(flyer, BECFlyerProtocol)


@pytest.mark.parametrize("center", [-10, 0, 10])
def test_monitor_readback(monitor, center):
    """Test the readback method of SimMonitor."""
    for model_name in monitor.sim.sim_get_models():
        monitor.sim.sim_select_model(model_name)
        monitor.sim.sim_params["noise_multipler"] = 10
        if "c" in monitor.sim.sim_params:
            monitor.sim.sim_params["c"] = center
        elif "center" in monitor.sim.sim_params:
            monitor.sim.sim_params["center"] = center
        assert isinstance(monitor.read()[monitor.name]["value"], monitor.BIT_DEPTH)
        expected_value = monitor.sim._model.eval(monitor.sim._model_params, x=0)
        print(expected_value, monitor.read()[monitor.name]["value"])
        tolerance = (
            monitor.sim.sim_params["noise_multipler"] + 1
        )  # due to ceiling in calculation, but maximum +1int
        assert np.isclose(
            monitor.read()[monitor.name]["value"],
            expected_value,
            atol=monitor.sim.sim_params["noise_multipler"] + 1,
        )


@pytest.mark.parametrize("amplitude, noise_multiplier", [(0, 1), (100, 10), (1000, 50)])
def test_camera_readback(camera, amplitude, noise_multiplier):
    """Test the readback method of SimMonitor."""
    for model_name in camera.sim.sim_get_models():
        camera.sim.sim_select_model(model_name)
        camera.sim.sim_params = {"noise_multiplier": noise_multiplier}
        camera.sim.sim_params = {"amplitude": amplitude}
        camera.sim.sim_params = {"noise": "poisson"}
        assert camera.image.get().shape == camera.SHAPE
        assert isinstance(camera.image.get()[0, 0], camera.BIT_DEPTH)
        camera.sim.sim_params = {"noise": "uniform"}
        camera.sim.sim_params = {"hot_pixel_coords": []}
        camera.sim.sim_params = {"hot_pixel_values": []}
        camera.sim.sim_params = {"hot_pixel_types": []}
        assert camera.image.get().shape == camera.SHAPE
        assert isinstance(camera.image.get()[0, 0], camera.BIT_DEPTH)
        assert (camera.image.get() <= (amplitude + noise_multiplier + 1)).all()


def test_positioner_move(positioner):
    """Test the move method of SimPositioner."""
    positioner.move(0).wait()
    assert np.isclose(positioner.read()[positioner.name]["value"], 0, atol=positioner.tolerance)
    positioner.move(10).wait()
    assert np.isclose(positioner.read()[positioner.name]["value"], 10, atol=positioner.tolerance)


@pytest.mark.parametrize("proxy_active", [True, False])
def test_sim_camera_proxies(camera, proxy_active):
    """Test mocking compute_method with framework class"""
    camera.device_manager.add_device("test_proxy")
    if proxy_active:
        camera._registered_proxies["test_proxy"] = camera.image.name
    else:
        camera._registered_proxies = {}
    proxy = camera.device_manager.devices["test_proxy"]
    mock_method = mock.MagicMock()
    mock_obj = proxy.obj
    mock_obj.lookup = mock.MagicMock()
    mock_obj.lookup.return_value = {camera.name: {"method": mock_method, "args": 1, "kwargs": 1}}
    camera.image.read()
    if proxy_active:
        assert len(mock_obj.lookup.mock_calls) > 0
    elif not proxy_active:
        assert len(mock_obj.lookup.mock_calls) == 0


def test_BECDeviceBase():
    # Test the BECDeviceBase class
    bec_device_base = BECDeviceBase(name="test")
    assert isinstance(bec_device_base, BECDevice)
    assert bec_device_base.connected is True
    signal = Signal(name="signal")
    assert isinstance(signal, BECDevice)
    device = Device(name="device")
    assert isinstance(device, BECDevice)


def test_h5proxy(h5proxy_fixture):
    """Test h5 camera proxy read from h5 file"""
    h5proxy, camera = h5proxy_fixture
    mock_proxy = mock.MagicMock()
    camera.device_manager.devices.update({h5proxy.name: mock_proxy})
    mock_proxy.enabled = True
    mock_proxy.obj = h5proxy
    fname = os.path.expanduser("tests/test_data/h5_test_file.h5")
    h5entry = "entry/data/data"
    with h5py.File(fname, "r") as f:
        data = f[h5entry][...]
    # pylint: disable=protected-access
    h5proxy._update_device_config(
        {camera.name: {"signal_name": "image", "file_source": fname, "h5_entry": h5entry}}
    )
    camera._registered_proxies.update({h5proxy.name: camera.image.name})
    camera.sim.sim_params = {"noise": "none", "noise_multiplier": 0}
    camera.scaninfo.sim_mode = True
    camera.stage()
    img = camera.image.get()
    assert (img == data[0, ...]).all()
    camera.unstage()


def test_slitproxy(slitproxy_fixture):
    """Test slit proxy to compute readback from readback of positioner samx"""
    proxy, camera, samx = slitproxy_fixture
    px_size = 0.5
    slitwidth = 2
    proxy._update_device_config(
        {
            camera.name: {
                "signal_name": "image",
                "center_offset": [0, 0],
                "covariance": [[1000, 500], [200, 1000]],
                "pixel_size": px_size,
                "ref_motors": [samx.name],
                "slit_width": [slitwidth],
                "motor_dir": [0],
            }
        }
    )
    camera._registered_proxies.update({proxy.name: camera.image.name})
    mock_proxy = mock.MagicMock()
    mock_samx = mock.MagicMock()
    mock_camera = mock.MagicMock()
    camera.device_manager.devices.update(
        {proxy.name: mock_proxy, samx.name: mock_samx, camera.name: mock_camera}
    )

    mock_proxy.enabled = True
    mock_samx.enabled = True
    mock_camera.enabled = True
    mock_camera.obj = camera
    mock_samx.obj = samx
    mock_proxy.obj = proxy
    camera.sim.sim_params = {"noise": "none", "noise_multiplier": 0, "hot_pixel_values": [0, 0, 0]}
    samx.delay = 0
    samx_pos = 0
    samx.move(samx_pos)
    proxy._gaussian_blur_sigma = 0
    img = camera.image.get()
    edges = (
        int(img.shape[0] // 2 - samx_pos / px_size - slitwidth / (2 * px_size)),
        int(img.shape[0] // 2 + samx_pos / px_size + slitwidth / (2 * px_size)),
    )
    assert (img[:, : edges[0]] == 0).all()
    assert (img[:, edges[1] :] == 0).all()
    samx_pos = 13.3
    samx.move(samx_pos)
    img = camera.image.get()
    edges = (
        int(img.shape[0] // 2 + samx_pos / px_size - slitwidth / (2 * px_size)),
        int(img.shape[0] // 2 + samx_pos / px_size + slitwidth / (2 * px_size)),
    )
    assert (img[:, : edges[0]] == 0).all()
    assert (img[:, edges[1] :] == 0).all()



import threading
import time

import numpy as np
from bec_lib import bec_logger
from ophyd import Component as Cpt
from ophyd import Device, PositionerBase, Signal
from ophyd.status import wait as status_wait
from ophyd.utils import LimitError

from ophyd_devices.galil.galil_ophyd import (
    BECConfigError,
    GalilAxesReferenced,
    GalilController,
    GalilError,
    GalilMotorIsMoving,
    GalilMotorResolution,
    GalilSetpointSignal,
    GalilSignalRO,
    retry_once,
)
from ophyd_devices.utils.controller import threadlocked
from ophyd_devices.utils.socket import SocketIO, raise_if_disconnected

logger = bec_logger.logger


class FlomniGalilController(GalilController):
    USER_ACCESS = [
        "describe",
        "show_running_threads",
        "galil_show_all",
        "socket_put_and_receive",
        "socket_put_confirmed",
        "drive_axis_to_limit",
        "find_reference",
        "get_motor_limit_switch",
        "fosaz_light_curtain_is_triggered",
        "is_motor_on",
        "all_axes_referenced",
        "lights_off",
        "lights_on",
    ]

    def is_axis_moving(self, axis_Id, axis_Id_numeric) -> bool:
        if axis_Id is None and axis_Id_numeric is not None:
            axis_Id = self.axis_Id_numeric_to_alpha(axis_Id_numeric)
        active_thread = self.is_thread_active(0)
        motor_is_on = self.is_motor_on(axis_Id)
        return bool(active_thread or motor_is_on)

    def all_axes_referenced(self) -> bool:
        # TODO: check if all axes are referenced in all controllers
        return super().all_axes_referenced()

    def fosaz_light_curtain_is_triggered(self) -> bool:
        """
        Check the light curtain status for fosaz

        Returns:
            bool: True if the light curtain is triggered
        """

        return int(float(self.socket_put_and_receive("MG @IN[14]").strip())) == 1

    def lights_off(self) -> None:
        """
        Turn off the lights
        """
        self.socket_put_confirmed("CB15")

    def lights_on(self) -> None:
        """
        Turn on the lights
        """
        self.socket_put_confirmed("SB15")


class FlomniGalilReadbackSignal(GalilSignalRO):
    @retry_once
    @threadlocked
    def _socket_get(self) -> float:
        """Get command for the readback signal

        Returns:
            float: Readback value after adjusting for sign and motor resolution.
        """

        current_pos = float(self.controller.socket_put_and_receive(f"TD{self.parent.axis_Id}"))
        current_pos *= self.parent.sign
        step_mm = self.parent.motor_resolution.get()
        return current_pos / step_mm

    def read(self):
        self._metadata["timestamp"] = time.time()
        val = super().read()
        return val


class FlomniGalilSetpointSignal(GalilSetpointSignal):
    @retry_once
    @threadlocked
    def _socket_set(self, val: float) -> None:
        """Set a new target value / setpoint value. Before submission, the target value is adjusted for the axis' sign.
        Furthermore, it is ensured that all axes are referenced before a new setpoint is submitted.

        Args:
            val (float): Target value / setpoint value

        Raises:
            GalilError: Raised if not all axes are referenced.

        """
        target_val = val * self.parent.sign
        self.setpoint = target_val
        axes_referenced = self.controller.all_axes_referenced()
        if axes_referenced:
            while self.controller.is_thread_active(0):
                time.sleep(0.1)

            self.controller.socket_put_confirmed(f"naxis={self.parent.axis_Id_numeric}")
            self.controller.socket_put_confirmed(f"ntarget={target_val:.3f}")
            self.controller.socket_put_confirmed("movereq=1")
            self.controller.socket_put_confirmed("XQ#NEWPAR")
            while self.controller.is_thread_active(0):
                time.sleep(0.005)
        else:
            raise GalilError("Not all axes are referenced.")


class FlomniGalilMotorResolution(GalilMotorResolution):
    pass


class FlomniGalilMotorIsMoving(GalilMotorIsMoving):
    pass


class FlomniGalilAxesReferenced(GalilAxesReferenced):
    pass


class FlomniGalilMotor(Device, PositionerBase):
    USER_ACCESS = ["controller"]
    readback = Cpt(FlomniGalilReadbackSignal, signal_name="readback", kind="hinted")
    user_setpoint = Cpt(FlomniGalilSetpointSignal, signal_name="setpoint")
    motor_resolution = Cpt(FlomniGalilMotorResolution, signal_name="resolution", kind="config")
    motor_is_moving = Cpt(FlomniGalilMotorIsMoving, signal_name="motor_is_moving", kind="normal")
    all_axes_referenced = Cpt(
        FlomniGalilAxesReferenced, signal_name="all_axes_referenced", kind="config"
    )
    high_limit_travel = Cpt(Signal, value=0, kind="omitted")
    low_limit_travel = Cpt(Signal, value=0, kind="omitted")

    SUB_READBACK = "readback"
    SUB_CONNECTION_CHANGE = "connection_change"
    _default_sub = SUB_READBACK

    def __init__(
        self,
        axis_Id,
        prefix="",
        *,
        name,
        kind=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        host="mpc2680.psi.ch",
        port=8081,
        limits=None,
        sign=1,
        socket_cls=SocketIO,
        device_manager=None,
        **kwargs,
    ):
        self.controller = FlomniGalilController(
            socket_cls=socket_cls, socket_host=host, socket_port=port
        )
        self.axis_Id = axis_Id
        self.controller.set_axis(axis=self, axis_nr=self.axis_Id_numeric)
        self.sign = sign
        self.tolerance = kwargs.pop("tolerance", 0.5)
        self.device_mapping = kwargs.pop("device_mapping", {})
        self.device_manager = device_manager

        if len(self.device_mapping) > 0 and self.device_manager is None:
            raise BECConfigError(
                "device_mapping has been specified but the device_manager cannot be accessed."
            )
        self.rt = self.device_mapping.get("rt")
        self.pid_x_correction = 0

        super().__init__(
            prefix,
            name=name,
            kind=kind,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            parent=parent,
            **kwargs,
        )
        self.readback.name = self.name
        self.controller.subscribe(
            self._update_connection_state, event_type=self.SUB_CONNECTION_CHANGE
        )
        self._update_connection_state()
        # self.readback.subscribe(self._forward_readback, event_type=self.readback.SUB_VALUE)

        if limits is not None:
            assert len(limits) == 2
            self.low_limit_travel.put(limits[0])
            self.high_limit_travel.put(limits[1])

    @property
    def limits(self):
        return (self.low_limit_travel.get(), self.high_limit_travel.get())

    @property
    def low_limit(self):
        return self.limits[0]

    @property
    def high_limit(self):
        return self.limits[1]

    def check_value(self, pos):
        """Check that the position is within the soft limits"""
        low_limit, high_limit = self.limits

        if low_limit < high_limit and not (low_limit <= pos <= high_limit):
            raise LimitError(f"position={pos} not within limits {self.limits}")

    def _update_connection_state(self, **kwargs):
        for walk in self.walk_signals():
            walk.item._metadata["connected"] = self.controller.connected

    def _forward_readback(self, **kwargs):
        kwargs.pop("sub_type")
        self._run_subs(sub_type="readback", **kwargs)

    @raise_if_disconnected
    def move(self, position, wait=True, **kwargs):
        """Move to a specified position, optionally waiting for motion to
        complete.

        Parameters
        ----------
        position
            Position to move to
        moved_cb : callable
            Call this callback when movement has finished. This callback must
            accept one keyword argument: 'obj' which will be set to this
            positioner instance.
        timeout : float, optional
            Maximum time to wait for the motion. If None, the default timeout
            for this positioner is used.

        Returns
        -------
        status : MoveStatus

        Raises
        ------
        TimeoutError
            When motion takes longer than `timeout`
        ValueError
            On invalid positions
        RuntimeError
            If motion fails other than timing out
        """
        self._started_moving = False
        timeout = kwargs.pop("timeout", 100)
        status = super().move(position, timeout=timeout, **kwargs)
        self.user_setpoint.put(position, wait=False)

        def move_and_finish():
            while self.motor_is_moving.get():
                logger.info("motor is moving")
                val = self.readback.read()
                self._run_subs(sub_type=self.SUB_READBACK, value=val, timestamp=time.time())
                time.sleep(0.1)
            val = self.readback.read()
            success = np.isclose(val[self.name]["value"], position, atol=self.tolerance)

            if not success:
                print(" stop")
            self._done_moving(success=success)
            logger.info("Move finished")

        threading.Thread(target=move_and_finish, daemon=True).start()
        try:
            if wait:
                status_wait(status)
        except KeyboardInterrupt:
            self.stop()
            raise

        return status

    @property
    def axis_Id(self):
        return self._axis_Id_alpha

    @axis_Id.setter
    def axis_Id(self, val):
        if isinstance(val, str):
            if len(val) != 1:
                raise ValueError("Only single-character axis_Ids are supported.")
            self._axis_Id_alpha = val
            self._axis_Id_numeric = ord(val.lower()) - 97
        else:
            raise TypeError(f"Expected value of type str but received {type(val)}")

    @property
    def axis_Id_numeric(self):
        return self._axis_Id_numeric

    @axis_Id_numeric.setter
    def axis_Id_numeric(self, val):
        if isinstance(val, int):
            if val > 26:
                raise ValueError("Numeric value exceeds supported range.")
            self._axis_Id_alpha = val
            self._axis_Id_numeric = (chr(val + 97)).capitalize()
        else:
            raise TypeError(f"Expected value of type int but received {type(val)}")

    @property
    def egu(self):
        """The engineering units (EGU) for positions"""
        return "mm"

    def stage(self) -> list[object]:
        return super().stage()

    def unstage(self) -> list[object]:
        return super().unstage()

    def stop(self, *, success=False):
        self.controller.stop_all_axes()
        return super().stop(success=success)


# if __name__ == "__main__":
#     mock = False
#     if not mock:
#         leyey = GalilMotor("H", name="leyey", host="mpc2680.psi.ch", port=8081, sign=-1)
#         leyey.stage()
#         status = leyey.move(0, wait=True)
#         status = leyey.move(10, wait=True)
#         leyey.read()

#         leyey.get()
#         leyey.describe()

#         leyey.unstage()
#     else:
#         from ophyd_devices.utils.socket import SocketMock

#         leyex = GalilMotor(
#             "G", name="leyex", host="mpc2680.psi.ch", port=8081, socket_cls=SocketMock
#         )
#         leyey = GalilMotor(
#             "H", name="leyey", host="mpc2680.psi.ch", port=8081, socket_cls=SocketMock
#         )
#         leyex.stage()
#         # leyey.stage()

#         leyex.controller.galil_show_all()



# pylint: skip-file
import threading
from unittest import mock

import ophyd
import pytest
from bec_lib import MessageEndpoints, messages

from ophyd_devices.epics.devices.mcs_csaxs import (
    MCScSAXS,
    MCSError,
    MCSTimeoutError,
    ReadoutMode,
    TriggerSource,
)
from tests.utils import DMMock, MockPV


def patch_dual_pvs(device):
    for walk in device.walk_signals():
        if not hasattr(walk.item, "_read_pv"):
            continue
        if not hasattr(walk.item, "_write_pv"):
            continue
        if walk.item._read_pv.pvname.endswith("_RBV"):
            walk.item._read_pv = walk.item._write_pv


@pytest.fixture(scope="function")
def mock_det():
    name = "mcs"
    prefix = "X12SA-MCS:"
    sim_mode = False
    dm = DMMock()
    with mock.patch.object(dm, "connector"):
        with (
            mock.patch("ophyd_devices.epics.devices.psi_detector_base.FileWriter") as filemixin,
            mock.patch(
                "ophyd_devices.epics.devices.psi_detector_base.PSIDetectorBase._update_service_config"
            ) as mock_service_config,
        ):
            with mock.patch.object(ophyd, "cl") as mock_cl:
                mock_cl.get_pv = MockPV
                mock_cl.thread_class = threading.Thread
                with mock.patch.object(MCScSAXS, "_init"):
                    det = MCScSAXS(name=name, prefix=prefix, device_manager=dm, sim_mode=sim_mode)
                    patch_dual_pvs(det)
                    yield det


def test_init():
    """Test the _init function:"""
    name = "eiger"
    prefix = "X12SA-ES-EIGER9M:"
    sim_mode = False
    dm = DMMock()
    with mock.patch.object(dm, "connector"):
        with (
            mock.patch("ophyd_devices.epics.devices.psi_detector_base.FileWriter"),
            mock.patch(
                "ophyd_devices.epics.devices.psi_detector_base.PSIDetectorBase._update_service_config"
            ),
        ):
            with mock.patch.object(ophyd, "cl") as mock_cl:
                mock_cl.get_pv = MockPV
                with (
                    mock.patch(
                        "ophyd_devices.epics.devices.mcs_csaxs.MCSSetup.initialize_default_parameter"
                    ) as mock_default,
                    mock.patch(
                        "ophyd_devices.epics.devices.mcs_csaxs.MCSSetup.initialize_detector"
                    ) as mock_init_det,
                    mock.patch(
                        "ophyd_devices.epics.devices.mcs_csaxs.MCSSetup.initialize_detector_backend"
                    ) as mock_init_backend,
                ):
                    MCScSAXS(name=name, prefix=prefix, device_manager=dm, sim_mode=sim_mode)
                    mock_default.assert_called_once()
                    mock_init_det.assert_called_once()
                    mock_init_backend.assert_called_once()


@pytest.mark.parametrize(
    "trigger_source, channel_advance, channel_source1, pv_channels",
    [
        (
            3,
            1,
            0,
            {
                "user_led": 0,
                "mux_output": 5,
                "input_pol": 0,
                "output_pol": 1,
                "count_on_start": 0,
                "stop_all": 1,
            },
        )
    ],
)
def test_initialize_detector(
    mock_det, trigger_source, channel_advance, channel_source1, pv_channels
):
    """Test the _init function:

    This includes testing the functions:
    - initialize_detector
    - stop_det
    - parent.set_trigger
    --> Testing the filewriter is done in test_init_filewriter

    Validation upon setting the correct PVs

    """
    mock_det.custom_prepare.initialize_detector()  # call the method you want to test
    assert mock_det.channel_advance.get() == channel_advance
    assert mock_det.channel1_source.get() == channel_source1
    assert mock_det.user_led.get() == pv_channels["user_led"]
    assert mock_det.mux_output.get() == pv_channels["mux_output"]
    assert mock_det.input_polarity.get() == pv_channels["input_pol"]
    assert mock_det.output_polarity.get() == pv_channels["output_pol"]
    assert mock_det.count_on_start.get() == pv_channels["count_on_start"]
    assert mock_det.input_mode.get() == trigger_source


def test_trigger(mock_det):
    """Test the trigger function:
    Validate that trigger calls the custom_prepare.on_trigger() function
    """
    with mock.patch.object(mock_det.custom_prepare, "on_trigger") as mock_on_trigger:
        mock_det.trigger()
        mock_on_trigger.assert_called_once()


@pytest.mark.parametrize(
    "value, num_lines, num_points, done", [(100, 5, 500, False), (500, 5, 500, True)]
)
def test_progress_update(mock_det, value, num_lines, num_points, done):
    mock_det.num_lines.set(num_lines)
    mock_det.scaninfo.num_points = num_points
    calls = mock.call(sub_type="progress", value=value, max_value=num_points, done=done)
    with mock.patch.object(mock_det, "_run_subs") as mock_run_subs:
        mock_det.custom_prepare._progress_update(value=value)
        mock_run_subs.assert_called_once()
        assert mock_run_subs.call_args == calls


@pytest.mark.parametrize(
    "values, expected_nothing",
    [([[100, 120, 140], [200, 220, 240], [300, 320, 340]], False), ([100, 200, 300], True)],
)
def test_on_mca_data(mock_det, values, expected_nothing):
    """Test the on_mca_data function:
    Validate that on_mca_data calls the custom_prepare.on_mca_data() function
    """
    with mock.patch.object(mock_det.custom_prepare, "_send_data_to_bec") as mock_send_data:
        mock_object = mock.MagicMock()
        for ii, name in enumerate(mock_det.custom_prepare.mca_names):
            mock_object.attr_name = name
            mock_det.custom_prepare._on_mca_data(obj=mock_object, value=values[ii])
            if not expected_nothing and ii < (len(values) - 1):
                assert mock_det.custom_prepare.mca_data[name] == values[ii]

        if not expected_nothing:
            mock_send_data.assert_called_once()
            assert mock_det.custom_prepare.acquisition_done is True


@pytest.mark.parametrize(
    "metadata, mca_data",
    [
        (
            {"scan_id": 123},
            {"mca1": [100, 120, 140], "mca3": [200, 220, 240], "mca4": [300, 320, 340]},
        )
    ],
)
def test_send_data_to_bec(mock_det, metadata, mca_data):
    mock_det.scaninfo.scan_msg = mock.MagicMock()
    mock_det.scaninfo.scan_msg.metadata = metadata
    mock_det.scaninfo.scan_id = metadata["scan_id"]
    mock_det.custom_prepare.mca_data = mca_data
    mock_det.custom_prepare._send_data_to_bec()
    device_metadata = mock_det.scaninfo.scan_msg.metadata
    metadata.update({"async_update": "append", "num_lines": mock_det.num_lines.get()})
    data = messages.DeviceMessage(signals=dict(mca_data), metadata=device_metadata)
    calls = mock.call(
        topic=MessageEndpoints.device_async_readback(
            scan_id=metadata["scan_id"], device=mock_det.name
        ),
        msg={"data": data},
        expire=1800,
    )

    assert mock_det.connector.xadd.call_args == calls


@pytest.mark.parametrize(
    "scaninfo, triggersource, stopped, expected_exception",
    [
        (
            {"num_points": 500, "frames_per_trigger": 1, "scan_type": "step"},
            TriggerSource.MODE3,
            False,
            False,
        ),
        (
            {"num_points": 500, "frames_per_trigger": 1, "scan_type": "fly"},
            TriggerSource.MODE3,
            False,
            False,
        ),
        (
            {"num_points": 5001, "frames_per_trigger": 2, "scan_type": "step"},
            TriggerSource.MODE3,
            False,
            True,
        ),
        (
            {"num_points": 500, "frames_per_trigger": 2, "scan_type": "random"},
            TriggerSource.MODE3,
            False,
            True,
        ),
    ],
)
def test_stage(mock_det, scaninfo, triggersource, stopped, expected_exception):
    mock_det.scaninfo.num_points = scaninfo["num_points"]
    mock_det.scaninfo.frames_per_trigger = scaninfo["frames_per_trigger"]
    mock_det.scaninfo.scan_type = scaninfo["scan_type"]
    mock_det.stopped = stopped
    with mock.patch.object(mock_det.custom_prepare, "prepare_detector_backend") as mock_prep_fw:
        if expected_exception:
            with pytest.raises(MCSError):
                mock_det.stage()
                mock_prep_fw.assert_called_once()
        else:
            mock_det.stage()
            mock_prep_fw.assert_called_once()
            # Check set_trigger
            mock_det.input_mode.get() == triggersource
            if scaninfo["scan_type"] == "step":
                assert mock_det.num_use_all.get() == int(scaninfo["frames_per_trigger"]) * int(
                    scaninfo["num_points"]
                )
            elif scaninfo["scan_type"] == "fly":
                assert mock_det.num_use_all.get() == int(scaninfo["num_points"])
            mock_det.preset_real.get() == 0

        # # CHeck custom_prepare.arm_acquisition
        # assert mock_det.custom_prepare.counter == 0
        # assert mock_det.erase_start.get() == 1
        # mock_prep_fw.assert_called_once()
        # # Check _prep_det
        # assert mock_det.cam.num_images.get() == int(
        #     scaninfo["num_points"] * scaninfo["frames_per_trigger"]
        # )
        # assert mock_det.cam.num_frames.get() == 1

        # mock_publish_file_location.assert_called_with(done=False)
        # assert mock_det.cam.acquire.get() == 1


def test_prepare_detector_backend(mock_det):
    mock_det.custom_prepare.prepare_detector_backend()
    assert mock_det.erase_all.get() == 1
    assert mock_det.read_mode.get() == ReadoutMode.EVENT


@pytest.mark.parametrize("stopped, expected_exception", [(False, False), (True, True)])
def test_unstage(mock_det, stopped, expected_exception):
    with (
        mock.patch.object(mock_det.custom_prepare, "finished") as mock_finished,
        mock.patch.object(
            mock_det.custom_prepare, "publish_file_location"
        ) as mock_publish_file_location,
    ):
        mock_det.stopped = stopped
        if expected_exception:
            mock_det.unstage()
            assert mock_det.stopped is True
        else:
            mock_det.unstage()
            mock_finished.assert_called_once()
            mock_publish_file_location.assert_called_with(done=True, successful=True)
            assert mock_det.stopped is False


def test_stop_detector_backend(mock_det):
    mock_det.custom_prepare.stop_detector_backend()
    assert mock_det.custom_prepare.acquisition_done is True


def test_stop(mock_det):
    with (
        mock.patch.object(mock_det.custom_prepare, "stop_detector") as mock_stop_det,
        mock.patch.object(
            mock_det.custom_prepare, "stop_detector_backend"
        ) as mock_stop_detector_backend,
    ):
        mock_det.stop()
        mock_stop_det.assert_called_once()
        mock_stop_detector_backend.assert_called_once()
        assert mock_det.stopped is True


@pytest.mark.parametrize(
    "stopped, acquisition_done, acquiring_state, expected_exception",
    [
        (False, True, 0, False),
        (False, False, 0, True),
        (False, True, 1, True),
        (True, True, 0, True),
    ],
)
def test_finished(mock_det, stopped, acquisition_done, acquiring_state, expected_exception):
    mock_det.custom_prepare.acquisition_done = acquisition_done
    mock_det.acquiring._read_pv.mock_data = acquiring_state
    mock_det.scaninfo.num_points = 500
    mock_det.num_lines.put(500)
    mock_det.current_channel._read_pv.mock_data = 1
    mock_det.stopped = stopped

    if expected_exception:
        with pytest.raises(MCSTimeoutError):
            mock_det.timeout = 0.1
            mock_det.custom_prepare.finished()
    else:
        mock_det.custom_prepare.finished()
        if stopped:
            assert mock_det.stopped is stopped



""" Module for Tomcat rotation motors.

The following classes implement the rotation motors for:

- AerotechAutomation1 (Tomcat), based on EpicsMotorIOC.

"""

import threading
import time

import numpy as np
from bec_lib import threadlocked
from ophyd import DeviceStatus

from ophyd_devices.ophyd_base_devices.bec_protocols import BECFlyerProtocol, BECScanProtocol
from ophyd_devices.ophyd_base_devices.ophyd_rotation_base import EpicsRotationBase


class TomcatAerotechRotation(EpicsRotationBase, BECFlyerProtocol, BECScanProtocol):
    """Special motor class that provides flyer interface and progress bar."""

    SUB_PROGRESS = "progress"

    def __init__(
        self,
        prefix="",
        *,
        name,
        kind=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        **kwargs,
    ):
        """Implementation of the Tomcat AerotechAutomation 1 rotation motor class.

        This motor class is based on EpicsRotationBase and provides in addition the flyer interface for BEC
        and a progress update.
        """
        super().__init__(
            prefix=prefix,
            name=name,
            kind=kind,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            parent=parent,
            **kwargs,
        )
        self._start_position = None
        self._target_position = None
        self._stopped = False
        self._rlock = threading.RLock()
        self.subscribe(self._progress_update, run=False)

    # ------------------ alternative to using configure method --------------------- #
    @property
    def start_position(self) -> float:
        """Get the start position."""
        return self._start_position

    @start_position.setter
    def start_position(self, value: float) -> None:
        """Set the start position."""
        self._start_position = value

    @property
    def target_position(self) -> float:
        """Get the start position."""
        return self._target_position

    @target_position.setter
    def target_position(self, value: float) -> None:
        """Set the start position."""
        self._target_position = value

    # ------------------ alternative to using configure method --------------------- #
    # configure method is optional for flyers within BEC, you can use stage method or pre_scan method to
    # set relevant parameters on the device.

    # def configure(self, d: dict) -> dict:
    #     """Configure method from the device.

    #     This method is usually used to set configuration parameters for the device.

    #     Args:
    #         d (dict): Dictionary with configuration parameters.

    #     """
    #     if "target" in d:
    #         self._target_position = d["target"]
    #         del d["target"]
    #     if "position" in d:
    #         self._target_position = d["position"]
    #         del d["position"]
    #     return super().configure(d)

    def pre_scan(self):
        """Perform pre-scan operation, e.g. move to start position."""
        if self._start_position:
            self.move(self._start_position, wait=True)

    def kickoff(self) -> DeviceStatus:
        """Kickoff the scan.

        The kickoff method should return a status object that is set to finish once the flyer flys, and is ready for the next actions.
        I would consider the following implementation.
        """
        self._start_position = float(self.position)
        self.move(self._target_position, wait=False)
        status = DeviceStatus(self)
        status.set_finished()
        return status

    def complete(self) -> DeviceStatus:
        """Complete method of the scan.

        This will be called in a fly scan after the kickoff, thus, the stage will be moving to it's target position.
        It should

        The complete method should return a status object that is set to finish once the flyer is done and the scan is complete.
        I would consider the following implementation.
        """
        threading.Thread(target=self._is_motor_moving, daemon=True).start()
        status = DeviceStatus(self)
        self.subscribe(status.set_finished, event_type=self.SUB_DONE, run=False)
        return status

    def stage(self) -> list[object]:
        """Stage the scan.

        We add here in addition the setting of the _stopped flag to False for the thread.
        """
        self._stopped = False
        return super().stage()

    def stop(self, success: bool = False) -> None:
        """Stop the scan.

        If the device is stopped, the _stopped flag is set to True.
        """
        self._stopped = True
        super().stop(success=success)

    @threadlocked
    def _is_motor_moving(self):
        """Function to check if the motor is moving.

        This function is used in a thread to check if the motor is moving.
        It resolves by running"""
        while self.motor_done_move.get():
            if self._stopped:
                self._done_moving(success=False)
                return
            time.sleep(0.1)
        self._done_moving(success=True)

    # TODO This logic could be refined to be more robust for various scan types, i.e. at the moment it just takes
    # the start and target position and calculates the progress based on the current position.
    def _progress_update(self, value, **kwargs) -> None:
        """Progress update on the scan.

        Runs the progress update on the device progress during the scan.
        It uses the SUB_PROGRESS event from ophyd to update BEC about the progress.
        Scans need to be aware which device progress is relevant for the scan.

        Args:
            value (float): The value of the motor position.
        """
        if (self._start_position is None) or (self._target_position is None) or (not self.moving):
            self._run_subs(sub_type=self.SUB_PROGRESS, value=1, max_value=1, done=1)
            return

        progress = np.abs(
            (value - self._start_position) / (self._target_position - self._start_position)
        )
        max_value = 100
        self._run_subs(
            sub_type=self.SUB_PROGRESS,
            value=int(100 * progress),
            max_value=max_value,
            done=int(np.isclose(max_value, progress, 1e-3)),
        )



# Standard ophyd classes
from ophyd import EpicsMotor, EpicsSignal, EpicsSignalRO
from ophyd.quadem import QuadEM
from ophyd.sim import SynAxis, SynPeriodicSignal, SynSignal

from .devices.delay_generator_csaxs import DelayGeneratorcSAXS
from .devices.flomni_sample_storage import FlomniSampleStorage
from .devices.InsertionDevice import InsertionDevice
from .devices.slits import SlitH, SlitV
from .devices.specMotors import (
    Bpm4i,
    EnergyKev,
    GirderMotorPITCH,
    GirderMotorROLL,
    GirderMotorX1,
    GirderMotorY1,
    GirderMotorYAW,
    MonoTheta1,
    MonoTheta2,
    PmDetectorRotation,
    PmMonoBender,
)
from .devices.SpmBase import SpmBase

# X07MA specific devices
from .devices.X07MADevices import *
from .devices.XbpmBase import XbpmBase, XbpmCsaxsOp



import abc
import functools
import logging
import socket
import time
import typing

import numpy as np
from bec_lib import bec_logger
from ophyd import Signal
from ophyd.utils.errors import DisconnectedError

logger = bec_logger.logger
# logger = bec_logger.logger("socket")


def raise_if_disconnected(fcn):
    """Decorator to catch attempted access to disconnected Galil channels."""

    @functools.wraps(fcn)
    def wrapper(self, *args, **kwargs):
        if self.connected:
            return fcn(self, *args, **kwargs)
        raise DisconnectedError(f"{self.name} is not connected")

    return wrapper


DEFAULT_EPICSSIGNAL_VALUE = object()

_type_map = {
    "number": (float, np.floating),
    "array": (np.ndarray, list, tuple),
    "string": (str,),
    "integer": (int, np.integer),
}


def data_shape(val):
    """Determine data-shape (dimensions)

    Returns
    -------
    list
        Empty list if val is number or string, otherwise
        ``list(np.ndarray.shape)``
    """
    if data_type(val) != "array":
        return []

    try:
        return list(val.shape)
    except AttributeError:
        return [len(val)]


def data_type(val):
    """Determine the JSON-friendly type name given a value

    Returns
    -------
    str
        One of {'number', 'integer', 'array', 'string'}

    Raises
    ------
    ValueError if the type is not recognized
    """
    bad_iterables = (str, bytes, dict)
    if isinstance(val, typing.Iterable) and not isinstance(val, bad_iterables):
        return "array"

    for json_type, py_types in _type_map.items():
        if isinstance(val, py_types):
            return json_type

    raise ValueError(
        f"Cannot determine the appropriate bluesky-friendly data type for "
        f"value {val} of Python type {type(val)}. "
        f"Supported types include: int, float, str, and iterables such as "
        f"list, tuple, np.ndarray, and so on."
    )


class SocketSignal(abc.ABC, Signal):
    SUB_SETPOINT = "setpoint"

    @abc.abstractmethod
    def _socket_get(self): ...

    @abc.abstractmethod
    def _socket_set(self, val): ...

    def get(self):
        self._readback = self._socket_get()
        return self._readback

    def put(
        self,
        value,
        force=False,
        connection_timeout=1,
        callback=None,
        use_complete=None,
        timeout=1,
        **kwargs,
    ):
        """Using channel access, set the write PV to `value`.

        Keyword arguments are passed on to callbacks

        Parameters
        ----------
        value : any
            The value to set
        force : bool, optional
            Skip checking the value in Python first
        connection_timeout : float, optional
            If not already connected, allow up to `connection_timeout` seconds
            for the connection to complete.
        use_complete : bool, optional
            Override put completion settings
        callback : callable
            Callback for when the put has completed
        timeout : float, optional
            Timeout before assuming that put has failed. (Only relevant if
            put completion is used.)
        """
        if not force:
            pass
            # self.check_value(value)

        self.wait_for_connection(timeout=connection_timeout)
        if use_complete is None:
            use_complete = False

        self._socket_set(value)
        old_value = self._parent.position

        timestamp = time.time()
        super().put(value, timestamp=timestamp, force=True)
        self._run_subs(
            sub_type=self.SUB_SETPOINT, old_value=old_value, value=value, timestamp=timestamp
        )

    def describe(self):
        """Provide schema and meta-data for :meth:`~BlueskyInterface.read`

        This keys in the `OrderedDict` this method returns must match the
        keys in the `OrderedDict` return by :meth:`~BlueskyInterface.read`.

        This provides schema related information, (ex shape, dtype), the
        source (ex PV name), and if available, units, limits, precision etc.

        Returns
        -------
        data_keys : OrderedDict
            The keys must be strings and the values must be dict-like
            with the ``event_model.event_descriptor.data_key`` schema.
        """
        if self._readback is DEFAULT_EPICSSIGNAL_VALUE:
            val = self.get()
        else:
            val = self._readback
        return {
            self.name: {
                "source": f"{self.parent.controller.name}:{self.name}",
                "dtype": data_type(val),
                "shape": data_shape(val),
            }
        }


class SocketIO:
    """SocketIO helper class for TCP IP connections"""

    def __init__(self, host, port, max_retry=10):
        self.host = host
        self.port = port
        self.is_open = False
        self.max_retry = max_retry
        self._initialize_socket()

    def connect(self):
        print(f"connecting to {self.host} port {self.port}")
        # self.sock.create_connection((host, port))
        retry_count = 0
        while True:
            try:
                if self.sock is None:
                    self._initialize_socket()
                self.sock.connect((self.host, self.port))
                break
            except Exception as exc:
                self.sock = None
                time.sleep(2)
                retry_count += 1
                if retry_count > self.max_retry:
                    raise exc

    def _put(self, msg_bytes):
        logger.debug(f"put message: {msg_bytes}")
        return self.sock.send(msg_bytes)

    def _recv(self, buffer_length=1024):
        msg = self.sock.recv(buffer_length)
        logger.debug(f"recv message: {msg}")
        return msg

    def _initialize_socket(self):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.settimeout(5)

    def put(self, msg):
        return self._put(msg)

    def receive(self, buffer_length=1024):
        return self._recv(buffer_length=buffer_length)

    def open(self):
        self.connect()
        self.is_open = True

    def close(self):
        self.sock.close()
        self.sock = None
        self.is_open = False


class SocketMock:
    def __init__(self, host, port):
        self.host = host
        self.port = port
        self.buffer_put = b""
        self.buffer_recv = [b" -12800"]
        self.is_open = False
        # self.open()

    def connect(self):
        print(f"connecting to {self.host} port {self.port}")

    def _put(self, msg_bytes):
        self.buffer_put = msg_bytes
        print(self.buffer_put)

    def _recv(self, buffer_length=1024):
        print(self.buffer_recv)
        if isinstance(self.buffer_recv, list):
            if len(self.buffer_recv) > 0:
                ret_val = self.buffer_recv.pop(0)
            else:
                ret_val = b""
            return ret_val
        return self.buffer_recv

    def _initialize_socket(self):
        pass

    def put(self, msg):
        return self._put(msg)

    def receive(self, buffer_length=1024):
        return self._recv(buffer_length=buffer_length)

    def open(self):
        self._initialize_socket()
        self.is_open = True

    def close(self):
        self.sock = None
        self.is_open = False



import enum
import os
import threading
import time as ttime

from bec_lib.logger import bec_logger

# from typing import Any
from ophyd import ADComponent as ADCpt
from ophyd import Component as Cpt
from ophyd import Device, EpicsSignal, EpicsSignalRO, EpicsSignalWithRBV

# import numpy as np
from ophyd.ophydobj import Kind

from ophyd_devices.epics.devices.psi_detector_base import CustomDetectorMixin, PSIDetectorBase

# os.environ["EPICS_CA_AUTO_ADDR_LIST"] = "No"
# os.environ["EPICS_CA_ADDR_LIST"] = "129.129.208.143"


logger = bec_logger.logger


class GrashopperError(Exception):
    """Base class for Grashopper detector errors."""


class GrashopperTimeoutError(GrashopperError):
    """Base class for Grashopper detector errors."""


class AutoMode(enum.IntEnum):
    """Acquire time auto for Grashopper detector.

    class for acquire_auto and gain_auto

    Off:        Gain tap balancing is user controlled using Gain.
    Once:       Gain tap balancing is automatically adjusted once by the device.
                Once it has converged, it automatically returns to the Off state.
    Continuous: Gain tap balancing is constantly adjusted by the device.
    """

    OFF = 0
    ONCE = 1
    CONTINUOUS = 2


class ImageMode(enum.IntEnum):
    """Image mode for Grashopper detector.

    Single: Acquire a single image, ignores NumImages PV
    Multiple: Acquire NumImages images
    Continuous: Acquire images continuously
    """

    SINGLE = 0
    MULTIPLE = 1
    CONTINUOUS = 2


class DetectorState(enum.IntEnum):
    """Detector states for Grashopper detector"""

    IDLE = 0
    ACQUIRE = 1
    READOUT = 2
    CORRECT = 3
    SAVING = 4
    ABORTING = 5
    ERROR = 6
    WAITING = 7
    INITIALIZING = 8
    DISCONNECTED = 9
    ABORTED = 10


class ImageBinning(enum.IntEnum):
    """Image binning for Grashopper detector"""

    X1 = 1
    X2 = 2
    X4 = 4


class VideoMode(enum.IntEnum):
    """Video mode for Grashopper detector.

    For details, consult EPICs IOC manual.
    """

    MODE0 = 0
    MODE1 = 1
    MODE2 = 2
    MODE3 = 3


class PixelFormat(enum.IntEnum):
    """Pixel format for Grashopper detector."""

    MONO8 = 0
    MONO12PACKED = 1
    MONO12P = 2
    MONO16 = 3


class COLORMODE(enum.IntEnum):
    """Color mode for Grashopper detector.

    Only for readback values from color_mode RO PV.
    """

    MONO = 0
    BAYER = 1
    RGB1 = 2
    RGB2 = 3
    RGB3 = 4
    YUV444 = 5
    YUV422 = 6
    YUV421 = 7


class TriggerSource(enum.IntEnum):
    """Trigger signals for Grashopper detector"""

    SOFTWARE = 0
    LINE0 = 1
    LINE2 = 2
    LINE3 = 3


class MemoryPolling(enum.IntEnum):
    """Memory polling for Grashopper detector.

    Defines update rate of memory polling for IOC (1s suggested).
    """

    PASSIVE = 0
    EVENT = 1
    IOINTR = 2
    SECONDS10 = 3
    SECONDS5 = 4
    SECONDS2 = 5
    SECONDS1 = 6
    SECONDS05 = 7
    SECONDS02 = 8
    SECONDS01 = 9


class GrashopperTOMCATSetup(CustomDetectorMixin):
    """Mixin class to setup TOMCAT specific implementations of the detector.

    This class will be called by the custom_prepare_cls attribute of the detector class.
    """

    def __init__(self, *_args, parent: Device = None, **_kwargs) -> None:
        super().__init__(*_args, parent=parent, **_kwargs)

        self.image_shape = (self.parent.cam.image_size_y.get(), self.parent.cam.image_size_x.get())
        self.monitor_thread = None
        self.stop_monitor = False
        self.update_frequency = 1
        self.low_frame_rate = 80

    def initialize_detector(self) -> None:
        """Initialize detector."""
        self.parent.cam.acquire.put(0)
        self.parent.cam.acquire_time_auto.put(AutoMode.CONTINUOUS)
        self.parent.cam.gain_auto.put(AutoMode.CONTINUOUS)
        self.parent.cam.image_mode.put(ImageMode.MULTIPLE)
        self.parent.cam.image_binning.put(ImageBinning.X1)
        self.parent.cam.video_mode.put(VideoMode.MODE0)
        self.parent.cam.pixel_format.put(PixelFormat.MONO16)
        self.parent.cam.trigger_source.put(TriggerSource.SOFTWARE)
        self.parent.cam.memory_polling.put(MemoryPolling.SECONDS1)
        self.parent.cam.set_image_counter.put(0)

    def initialize_detector_backend(self) -> None:
        self.parent.image.queue_size.put(5)
        self.parent.image.array_port.put(self.parent.cam.port_name.get())
        self.parent.image.enable_cb.put(1)
        self.parent.image.set_array_counter.put(0)

    def set_exposure_time(self, exposure_time: float) -> None:
        """Set the detector framerate.

        Args:
            framerate (float): Desired framerate in Hz smallest is 87Hz
        """
        framerate = 1 / exposure_time
        if framerate > self.low_frame_rate:
            raise GrashopperError(
                f"Trying to set exposure time to {exposure_time}s, this is below the lowest"
                f" possible exposure of {1/self.low_frame_rate}s"
            )
        self.parent.cam.frame_rate.put(framerate)

    def prepare_detector(self) -> None:
        """Prepare detector for acquisition."""
        self.parent.cam.image_mode.put(ImageMode.MULTIPLE)
        self.parent.cam.acquire_time_auto.put(AutoMode.CONTINUOUS)
        self.set_exposure_time(self.parent.scaninfo.exp_time)
        self.parent.set_trigger(TriggerSource.SOFTWARE)
        self.parent.cam.set_image_counter.put(0)
        self.set_acquisition_params()

    def set_acquisition_params(self) -> None:
        """Set acquisition parameters for the detector"""

        # Set number of images and frames (frames is for internal burst of detector)
        self.parent.cam.num_images.put(
            int(self.parent.scaninfo.num_points * self.parent.scaninfo.frames_per_trigger)
        )

    def prepare_detector_backend(self) -> None:
        """Prepare detector backend for acquisition."""
        self.parent.image.set_array_counter.put(0)
        self.monitor_thread = None
        self.stop_monitor = False
        # self.run_monitor()

    def arm_acquisition(self) -> None:
        """Arm grashopper detector for acquisition"""
        self.parent.cam.acquire.put(1)
        signal_conditions = [(self.parent.cam.detector_state.get, DetectorState.WAITING)]
        if not self.wait_for_signals(
            signal_conditions=signal_conditions,
            timeout=self.parent.timeout,
            check_stopped=True,
            all_signals=False,
        ):
            raise GrashopperTimeoutError(
                f"Failed to arm the acquisition. Detector state {signal_conditions[0][0]}"
            )

    def on_trigger(self) -> None:
        """Trigger the detector"""
        if self.parent.cam.trigger_source.get() == TriggerSource.SOFTWARE:
            self.parent.cam.software_trigger_device.put(1)
            ttime.sleep(0.1)
            self.send_data()

    def run_monitor(self) -> None:
        """
        Run the monitor loop in a separate thread.
        """
        self.monitor_thread = threading.Thread(target=self.monitor_loop, daemon=True)
        self.monitor_thread.start()

    def monitor_loop(self) -> None:
        """
        Monitor the detector status and send data.
        """
        while True:
            self.send_data()
            ttime.sleep(1 / self.update_frequency)
            if self.parent.stopped:
                break

    def send_data(self) -> None:
        """Send data to monitor endpoint in redis."""
        try:
            img = self.parent.image.array_data.get().reshape(self.image_shape)
            # pylint: disable=protected-access
            self.parent._run_subs(sub_type=self.parent.SUB_VALUE, value=img)
        except Exception as e:
            logger.debug(f"{e} for image with shape {self.parent.image.array_data.get().shape}")

    def stop_detector(self) -> None:
        """Stop detector."""
        self.parent.cam.acquire.put(0)
        signal_conditions = [(self.parent.cam.detector_state.get, DetectorState.IDLE)]
        if not self.wait_for_signals(
            signal_conditions=signal_conditions,
            timeout=self.parent.timeout - self.parent.timeout // 2,
            check_stopped=True,
            all_signals=False,
        ):
            # Retry stop detector and wait for remaining time
            self.parent.cam.acquire.put(0)
            if not self.wait_for_signals(
                signal_conditions=signal_conditions,
                timeout=self.parent.timeout - self.parent.timeout // 2,
                check_stopped=True,
                all_signals=False,
            ):
                raise GrashopperTimeoutError(
                    f"Failed to stop detector, detector state {signal_conditions[0][0]}"
                )

    def stop_detector_backend(self) -> None:
        """Stop the data backend sending data."""
        self.stop_monitor = True


class SLSDetectorCam(Device):
    """
    SLS Detector Camera - Grashoppter

    Base class to map EPICS PVs to ophyd signals.
    """

    ## Deprecated PVs, to be checked!
    # acquire_time = ADCpt(EpicsSignal, "AcquireTime", kind=Kind.omitted)
    # num_exposures = ADCpt(EpicsSignal, "NumExposures", kind=Kind.omitted)
    # acquire_period = ADCpt(EpicsSignalWithRBV, "AcquirePeriod", kind=Kind.config)

    # Control PVs
    acquire_time_auto = ADCpt(EpicsSignal, "AcquireTimeAuto", kind=Kind.config)

    frame_rate = ADCpt(EpicsSignalWithRBV, "FrameRate", kind=Kind.normal)
    num_images = ADCpt(EpicsSignalWithRBV, "NumImages", kind=Kind.normal)
    num_images_counter = ADCpt(EpicsSignalRO, "NumImagesCounter_RBV", kind=Kind.normal)
    image_mode = ADCpt(EpicsSignalWithRBV, "ImageMode", kind=Kind.config)
    acquire = ADCpt(EpicsSignalWithRBV, "Acquire", kind=Kind.config)
    detector_state = ADCpt(EpicsSignalRO, "DetectorState_RBV", kind=Kind.normal)
    status_message = ADCpt(EpicsSignalRO, "StatusMessage_RBV", string=True, kind=Kind.config)

    set_image_counter = ADCpt(EpicsSignal, "ArrayCounter", kind=Kind.config)
    image_counter = ADCpt(EpicsSignal, "ArrayCounter_RBV", kind=Kind.normal)
    image_rate = ADCpt(EpicsSignalRO, "ArrayRate_RBV", kind=Kind.omitted)

    # Asyn Port name
    port_name = ADCpt(EpicsSignalRO, "PortName_RBV", string=True, kind=Kind.omitted)

    # Readout related PVs
    max_image_size_x = ADCpt(EpicsSignalRO, "MaxSizeX_RBV", kind=Kind.config)
    max_image_size_y = ADCpt(EpicsSignalRO, "MaxSizeY_RBV", kind=Kind.config)
    image_size_x = ADCpt(EpicsSignalRO, "ArraySizeX_RBV", kind=Kind.config)
    image_size_y = ADCpt(EpicsSignalRO, "ArraySizeY_RBV", kind=Kind.config)
    # Only BinY PV is working, sets both
    image_binning = ADCpt(EpicsSignalWithRBV, "BinY", kind=Kind.config)

    gain = ADCpt(EpicsSignalWithRBV, "Gain", kind=Kind.config)
    gain_auto = ADCpt(EpicsSignalWithRBV, "GainAuto", kind=Kind.config)
    video_mode = ADCpt(EpicsSignalWithRBV, "VideoMode", kind=Kind.config)
    pixel_format = ADCpt(EpicsSignalWithRBV, "PixelFormat", kind=Kind.config)
    # Desired to set this in future?
    color_mode = ADCpt(EpicsSignalRO, "ColorMode_RBV", kind=Kind.config)

    # HW Status PVs
    temperature_actual = ADCpt(EpicsSignal, "TemperatureActual", kind=Kind.omitted)

    # Trigger
    trigger_mode_active = ADCpt(EpicsSignalWithRBV, "TriggerMode", kind=Kind.config)
    trigger_source = ADCpt(EpicsSignalWithRBV, "TriggerSource", kind=Kind.config)
    trigger_delay = ADCpt(EpicsSignalWithRBV, "TriggerDelay", kind=Kind.omitted)
    exposure_mode = ADCpt(EpicsSignalWithRBV, "ExposureMode", kind=Kind.omitted)
    software_trigger_device = ADCpt(EpicsSignal, "SoftwareTrigger", kind=Kind.config)

    # buffer
    memory_polling = ADCpt(EpicsSignal, "PoolUsedMem.SCAN", kind=Kind.omitted)


class SLSImagePlugin(Device):
    """SLS Image Plugin

    Image plugin for SLS detector imitating the behaviour of ImagePlugin from
    ophyd's areadetector plugins.
    """

    # Control
    array_port = Cpt(EpicsSignal, "NDArrayPort", kind=Kind.omitted, string=True)
    enable_cb = Cpt(EpicsSignal, "EnableCallbacks", kind=Kind.config)
    queue_size = Cpt(EpicsSignal, "QueueSize", kind=Kind.config)
    set_array_counter = Cpt(EpicsSignal, "ArrayCounter", kind=Kind.config)
    array_counter = Cpt(EpicsSignal, "ArrayCounter_RBV", kind=Kind.normal)
    set_dropped_arrays = Cpt(EpicsSignal, "DroppedArrays", kind=Kind.config)
    dropped_arrays = Cpt(EpicsSignal, "DroppedArrays_RBV", kind=Kind.normal)
    image_id = Cpt(EpicsSignal, "UniqueId_RBV", kind=Kind.normal)

    # Data
    array_data = Cpt(EpicsSignal, "ArrayData", kind=Kind.omitted)

    # Size related PVs from Plugin
    array_size_0 = Cpt(EpicsSignalRO, "ArraySize0_RBV", kind=Kind.omitted)
    array_size_1 = Cpt(EpicsSignalRO, "ArraySize1_RBV", kind=Kind.omitted)
    array_size_2 = Cpt(EpicsSignalRO, "ArraySize2_RBV", kind=Kind.omitted)
    array_dimension_size = Cpt(EpicsSignalRO, "NDimensions_RBV", kind=Kind.omitted)


class GrashopperTOMCAT(PSIDetectorBase):
    """
    Grashopper detector for TOMCAT

    Parent class: PSIDetectorBase

    class attributes:
        custom_prepare_cls (GrashopperTOMCATSetup)        : Custom detector setup class for TOMCAT,
                                                            inherits from CustomDetectorMixin
        cam (SLSDetectorCam)                              : Detector camera
        image (SLSImagePlugin)                            : Image plugin for detector
    """

    # Specify which functions are revealed to the user in BEC client
    USER_ACCESS = ["describe"]

    SUB_MONITOR = "monitor"
    SUB_VALUE = "value"
    _default_sub = SUB_VALUE

    # specify Setup class
    custom_prepare_cls = GrashopperTOMCATSetup
    # specify minimum readout time for detector
    MIN_READOUT = 0
    # specify class attributes
    cam = ADCpt(SLSDetectorCam, "cam1:")
    image = ADCpt(SLSImagePlugin, "image1:")

    def stage(self) -> list[object]:
        rtr = super().stage()
        self.custom_prepare.arm_acquisition()
        return rtr

    def unstage(self) -> list[object]:
        rtr = super().unstage()
        self.custom_prepare.stop_monitor = True
        return rtr

    def set_trigger(self, trigger_source: TriggerSource) -> None:
        """Set trigger source for the detector.
        Check the TriggerSource enum for possible values

        Args:
            trigger_source (TriggerSource): Trigger source for the detector

        """
        value = trigger_source
        self.cam.trigger_source.put(value)


if __name__ == "__main__":
    hopper = GrashopperTOMCAT(name="hopper", prefix="X02DA-PG-USB:", sim_mode=True)
    hopper.wait_for_connection(all_signals=True)



import functools
import threading

from bec_lib import bec_logger
from ophyd import Device
from ophyd.ophydobj import OphydObject

logger = bec_logger.logger


def threadlocked(fcn):
    """Ensure that the thread acquires and releases the lock."""

    @functools.wraps(fcn)
    def wrapper(self, *args, **kwargs):
        lock = self._lock if hasattr(self, "_lock") else self.controller._lock
        with lock:
            return fcn(self, *args, **kwargs)

    return wrapper


def axis_checked(fcn):
    """Decorator to catch attempted access to channels that are not available."""

    @functools.wraps(fcn)
    def wrapper(self, *args, **kwargs):
        if "axis_nr" in kwargs:
            self._check_axis_number(kwargs["axis_nr"])
        elif "axis_Id_numeric" in kwargs:
            self._check_axis_number(kwargs["axis_Id_numeric"])
        elif args:
            self._check_axis_number(args[0])
        return fcn(self, *args, **kwargs)

    return wrapper


class Controller(OphydObject):
    """Base class for all socker-based controllers."""

    _controller_instances = {}
    _initialized = False
    _axes_per_controller = 1

    SUB_CONNECTION_CHANGE = "connection_change"

    def __init__(
        self,
        *,
        name=None,
        socket_cls=None,
        socket_host=None,
        socket_port=None,
        attr_name="",
        parent=None,
        labels=None,
        kind=None,
    ):
        if not self._initialized:
            super().__init__(
                name=name, attr_name=attr_name, parent=parent, labels=labels, kind=kind
            )
            self._lock = threading.RLock()
            self._axis = []
            self._initialize()
            self._initialized = True
            self.sock = None
            self._socket_cls = socket_cls
            self._socket_host = socket_host
            self._socket_port = socket_port

    def _initialize(self):
        self._connected = False
        self._set_default_values()

    def _set_default_values(self):
        # no. of axes controlled by each controller
        self._axis = [None for axis_num in range(self._axes_per_controller)]

    @classmethod
    def _reset_controller(cls):
        cls._controller_instances = {}
        cls._initialized = False

    @property
    def connected(self):
        return self._connected

    @connected.setter
    def connected(self, value):
        self._connected = value
        self._run_subs(sub_type=self.SUB_CONNECTION_CHANGE)

    @axis_checked
    def set_axis(self, *, axis: Device, axis_nr: int) -> None:
        """Assign an axis to a device instance.

        Args:
            axis (Device): Device instance (e.g. GalilMotor)
            axis_nr (int): Controller axis number

        """
        self._axis[axis_nr] = axis

    @axis_checked
    def get_axis(self, axis_nr: int) -> Device:
        """Get device instance for a specified controller axis.

        Args:
            axis_nr (int): Controller axis number

        Returns:
            Device: Device instance (e.g. GalilMotor)

        """
        return self._axis[axis_nr]

    def _check_axis_number(self, axis_Id_numeric: int) -> None:
        if axis_Id_numeric >= self._axes_per_controller:
            raise ValueError(
                f"Axis {axis_Id_numeric} exceeds the available number of axes ({self._axes_per_controller})"
            )

    def on(self) -> None:
        """Open a new socket connection to the controller"""
        if not self.connected or self.sock is None:
            self.sock = self._socket_cls(host=self._socket_host, port=self._socket_port)
            self.sock.open()
            self.connected = True
        else:
            logger.info("The connection has already been established.")

    def off(self) -> None:
        """Close the socket connection to the controller"""
        if self.connected and self.sock is not None:
            self.sock.close()
            self.connected = False
            self.sock = None
        else:
            logger.info("The connection is already closed.")

    def __new__(cls, *args, **kwargs):
        socket_cls = kwargs.get("socket_cls")
        socket_host = kwargs.get("socket_host")
        socket_port = kwargs.get("socket_port")
        if not socket_cls:
            raise RuntimeError("Socket class must be specified.")
        if not socket_host:
            raise RuntimeError("Socket host must be specified.")
        if not socket_port:
            raise RuntimeError("Socket port must be specified.")
        host_port = f"{socket_host}:{socket_port}"
        if host_port not in cls._controller_instances:
            cls._controller_instances[host_port] = object.__new__(cls)
        return cls._controller_instances[host_port]



# pylint: skip-file
from unittest import mock

import pytest

from ophyd_devices.epics.devices.grashopper_tomcat import (
    COLORMODE,
    AutoMode,
    DetectorState,
    GrashopperError,
    GrashopperTimeoutError,
    GrashopperTOMCATSetup,
    ImageBinning,
    ImageMode,
    MemoryPolling,
    PixelFormat,
    TriggerSource,
    VideoMode,
)


def patch_dual_pvs(device):
    for walk in device.walk_signals():
        if not hasattr(walk.item, "_read_pv"):
            continue
        if not hasattr(walk.item, "_write_pv"):
            continue
        if walk.item._read_pv.pvname.endswith("_RBV"):
            walk.item._read_pv = walk.item._write_pv


@pytest.fixture(scope="function")
def mock_GrashopperSetup():
    mock_Grashopper = mock.MagicMock()
    yield GrashopperTOMCATSetup(parent=mock_Grashopper)


# Fixture for scaninfo
@pytest.fixture(
    params=[
        {
            "scan_id": "1234",
            "scan_type": "step",
            "num_points": 500,
            "frames_per_trigger": 1,
            "exp_time": 0.1,
            "readout_time": 0.1,
        },
        {
            "scan_id": "1234",
            "scan_type": "step",
            "num_points": 500,
            "frames_per_trigger": 5,
            "exp_time": 0.01,
            "readout_time": 0,
        },
        {
            "scan_id": "1234",
            "scan_type": "fly",
            "num_points": 500,
            "frames_per_trigger": 1,
            "exp_time": 1,
            "readout_time": 0.2,
        },
        {
            "scan_id": "1234",
            "scan_type": "fly",
            "num_points": 500,
            "frames_per_trigger": 5,
            "exp_time": 0.1,
            "readout_time": 0.4,
        },
    ]
)
def scaninfo(request):
    return request.param


def test_initialize_detector(mock_GrashopperSetup):
    """Test the initialize_detector method."""
    # Call the function you want to test
    mock_GrashopperSetup.initialize_detector()

    # Assert the correct methods are called with the expected arguments
    mock_GrashopperSetup.parent.cam.acquire.put.assert_called_once_with(0)
    mock_GrashopperSetup.parent.cam.acquire_time_auto.put.assert_called_once_with(
        AutoMode.CONTINUOUS
    )
    mock_GrashopperSetup.parent.cam.gain_auto.put.assert_called_once_with(AutoMode.CONTINUOUS)
    mock_GrashopperSetup.parent.cam.image_mode.put.assert_called_once_with(ImageMode.MULTIPLE)
    mock_GrashopperSetup.parent.cam.image_binning.put.assert_called_once_with(ImageBinning.X1)
    mock_GrashopperSetup.parent.cam.video_mode.put.assert_called_once_with(VideoMode.MODE0)
    mock_GrashopperSetup.parent.cam.pixel_format.put.assert_called_once_with(PixelFormat.MONO16)
    mock_GrashopperSetup.parent.cam.trigger_source.put.assert_called_once_with(
        TriggerSource.SOFTWARE
    )
    mock_GrashopperSetup.parent.cam.memory_polling.put.assert_called_once_with(
        MemoryPolling.SECONDS1
    )
    mock_GrashopperSetup.parent.cam.set_image_counter.put.assert_called_once_with(0)


def test_initialize_detector_backend(mock_GrashopperSetup):
    """Test the initialize_detector_backend method."""
    # Call the function you want to test
    mock_GrashopperSetup.initialize_detector_backend()

    # Assert the correct methods are called with the expected arguments
    mock_GrashopperSetup.parent.image.queue_size.put.assert_called_once_with(5)
    mock_GrashopperSetup.parent.image.array_port.put.assert_called_once_with(
        mock_GrashopperSetup.parent.cam.port_name.get()
    )
    mock_GrashopperSetup.parent.image.enable_cb.put.assert_called_once_with(1)
    mock_GrashopperSetup.parent.image.set_array_counter.put.assert_called_once_with(0)


@pytest.mark.parametrize("exposure_time", [1, 0.1, 0.01, 0.001, 0.0001])
def test_set_exposure_time(mock_GrashopperSetup, exposure_time):
    """Test the set_exposure_time method."""
    # Call the function you want to test
    frame_rate = 1 / exposure_time
    if frame_rate > mock_GrashopperSetup.low_frame_rate:
        with pytest.raises(GrashopperError):
            mock_GrashopperSetup.set_exposure_time(exposure_time)
    else:
        mock_GrashopperSetup.set_exposure_time(exposure_time)
        mock_GrashopperSetup.parent.cam.frame_rate.put.assert_called_once_with(frame_rate)


def test_prepare_detector(mock_GrashopperSetup, scaninfo):
    """Test the prepare_detector method."""
    # setup scaninof
    for k, v in scaninfo.items():
        setattr(mock_GrashopperSetup.parent.scaninfo, k, v)

    # Call the function you want to test
    with (
        mock.patch.object(
            mock_GrashopperSetup, "set_acquisition_params"
        ) as mock_set_acquisition_params,
        mock.patch.object(mock_GrashopperSetup, "set_exposure_time") as mock_set_exposure_time,
    ):
        mock_GrashopperSetup.prepare_detector()

        # Assert the correct methods are called with the expected arguments
        mock_GrashopperSetup.parent.cam.image_mode.put.assert_called_once_with(ImageMode.MULTIPLE)
        mock_GrashopperSetup.parent.cam.acquire_time_auto.put.assert_called_once_with(
            AutoMode.CONTINUOUS
        )
        mock_GrashopperSetup.parent.set_trigger.assert_called_once_with(TriggerSource.SOFTWARE)
        mock_GrashopperSetup.parent.cam.set_image_counter.put.assert_called_once_with(0)
        mock_set_acquisition_params.assert_called_once()
        mock_set_exposure_time.assert_called_once_with(scaninfo["exp_time"])


def test_prepare_detector_backend(mock_GrashopperSetup):
    """Test the prepare_detector_backend method."""
    # Call the function you want to test
    mock_GrashopperSetup.prepare_detector_backend()
    mock_GrashopperSetup.parent.image.set_array_counter.put.assert_called_once_with(0)
    assert mock_GrashopperSetup.monitor_thread is None
    assert not mock_GrashopperSetup.stop_monitor


@pytest.mark.parametrize("detector_state", [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
def test_arm_acquisition(mock_GrashopperSetup, detector_state):
    """Test the arm_acquisition method."""
    # Call the function you want to test
    mock_GrashopperSetup.parent.cam.detector_state.get.return_value = detector_state
    mock_GrashopperSetup.parent.timeout = 0.1

    if detector_state != DetectorState.WAITING:
        with pytest.raises(GrashopperTimeoutError):
            mock_GrashopperSetup.arm_acquisition()
            assert mock_GrashopperSetup.parent.cam.acquire.put.call_count == 1
            assert mock_GrashopperSetup.parent.cam.acquire.put.call_args == mock.call(1)
    else:
        mock_GrashopperSetup.arm_acquisition()
        assert mock_GrashopperSetup.parent.cam.acquire.put.call_count == 1
        assert mock_GrashopperSetup.parent.cam.acquire.put.call_args == mock.call(1)


@pytest.mark.parametrize("trigger_source", [0, 1, 2, 3])
def test_on_trigger(mock_GrashopperSetup, trigger_source):
    """Test the on_trigger method."""
    # Call the function you want to test
    mock_GrashopperSetup.parent.cam.trigger_source.get.return_value = trigger_source
    with mock.patch.object(mock_GrashopperSetup, "send_data") as mock_send_data:
        mock_GrashopperSetup.on_trigger()

    # Assert the correct methods are called with the expected arguments
    if trigger_source == TriggerSource.SOFTWARE:
        mock_GrashopperSetup.parent.cam.software_trigger_device.put.assert_called_once_with(1)
        assert mock_send_data.call_count == 1


def test_set_acquisition_params(mock_GrashopperSetup, scaninfo):
    """Test the set_acquisition_params method."""
    # Setup scaninfo
    mock_GrashopperSetup.parent.scaninfo.num_points = scaninfo["num_points"]
    mock_GrashopperSetup.parent.scaninfo.frames_per_trigger = scaninfo["frames_per_trigger"]

    # Call the function you want to test
    mock_GrashopperSetup.set_acquisition_params()

    # Assert the correct methods are called with the expected arguments
    expected_num_images = scaninfo["num_points"] * scaninfo["frames_per_trigger"]
    mock_GrashopperSetup.parent.cam.num_images.put.assert_called_once_with(expected_num_images)


@pytest.mark.parametrize("detector_state", [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
def test_stop_detector(mock_GrashopperSetup, detector_state):
    """Test the arm_acquisition method."""
    # Call the function you want to test
    mock_GrashopperSetup.parent.cam.detector_state.get.return_value = detector_state
    mock_GrashopperSetup.parent.timeout = 0.1

    if detector_state != DetectorState.IDLE:
        with pytest.raises(GrashopperTimeoutError):
            mock_GrashopperSetup.stop_detector()
            assert mock_GrashopperSetup.parent.cam.acquire.put.call_count == 2
            assert mock_GrashopperSetup.parent.cam.acquire.put.call_args == mock.call(0)
    else:
        mock_GrashopperSetup.stop_detector()
        assert mock_GrashopperSetup.parent.cam.acquire.put.call_count == 1
        assert mock_GrashopperSetup.parent.cam.acquire.put.call_args == mock.call(0)



from unittest import mock

import pytest
from utils import SocketMock

from ophyd_devices.rt_lamni import RtFlomniController, RtFlomniMotor
from ophyd_devices.rt_lamni.rt_ophyd import RtError


@pytest.fixture()
def rt_flomni():
    rt_flomni = RtFlomniController(
        name="rt_flomni", socket_cls=SocketMock, socket_host="localhost", socket_port=8081
    )
    with mock.patch.object(rt_flomni, "get_device_manager"):
        with mock.patch.object(rt_flomni, "sock"):
            rtx = mock.MagicMock(spec=RtFlomniMotor)
            rtx.name = "rtx"
            rty = mock.MagicMock(spec=RtFlomniMotor)
            rty.name = "rty"
            rtz = mock.MagicMock(spec=RtFlomniMotor)
            rtz.name = "rtz"
            rt_flomni.set_axis(rtx, 0)
            rt_flomni.set_axis(rty, 1)
            rt_flomni.set_axis(rtz, 2)
            yield rt_flomni


def test_rt_flomni_move_to_zero(rt_flomni):
    rt_flomni.move_to_zero()
    assert rt_flomni.sock.mock_calls == [
        mock.call.put(b"pa0,0\n"),
        mock.call.put(b"pa1,0\n"),
        mock.call.put(b"pa2,0\n"),
    ]


@pytest.mark.parametrize("return_value,is_running", [(b"1.00\n", False), (b"0.00\n", True)])
def test_rt_flomni_feedback_is_running(rt_flomni, return_value, is_running):
    rt_flomni.sock.receive.return_value = return_value
    assert rt_flomni.feedback_is_running() == is_running
    assert mock.call.put(b"l2\n") in rt_flomni.sock.mock_calls


def test_feedback_enable_with_reset(rt_flomni):

    device_manager = rt_flomni.get_device_manager()
    device_manager.devices.fsamx.user_parameter.get.return_value = 0.05
    device_manager.devices.fsamx.obj.readback.get.return_value = 0.05

    with mock.patch.object(rt_flomni, "feedback_is_running", return_value=True):
        with mock.patch.object(rt_flomni, "laser_tracker_on") as laser_tracker_on:
            with mock.patch.object(rt_flomni, "pid_y", return_value=0.05):
                with mock.patch.object(
                    rt_flomni, "slew_rate_limiters_on_target", return_value=True
                ) as slew_rate_limiters_on_target:

                    rt_flomni.feedback_enable_with_reset()
                    laser_tracker_on.assert_called_once()


def test_move_samx_to_scan_region(rt_flomni):
    device_manager = rt_flomni.get_device_manager()
    device_manager.devices.rtx.user_parameter.get.return_value = 1
    rt_flomni.move_samx_to_scan_region(20, 2)
    assert mock.call(b"v0\n") not in rt_flomni.sock.put.mock_calls
    assert mock.call(b"v1\n") in rt_flomni.sock.put.mock_calls


def test_feedback_enable_without_reset(rt_flomni):
    with mock.patch.object(rt_flomni, "set_device_enabled") as set_device_enabled:
        with mock.patch.object(rt_flomni, "feedback_is_running", return_value=True):
            with mock.patch.object(rt_flomni, "laser_tracker_on") as laser_tracker_on:
                rt_flomni.feedback_enable_without_reset()
                laser_tracker_on.assert_called_once()
                assert mock.call(b"l3\n") in rt_flomni.sock.put.mock_calls
                assert mock.call("fsamx", False) in set_device_enabled.mock_calls
                assert mock.call("fsamy", False) in set_device_enabled.mock_calls
                assert mock.call("foptx", False) in set_device_enabled.mock_calls
                assert mock.call("fopty", False) in set_device_enabled.mock_calls


def test_feedback_enable_without_reset_raises(rt_flomni):
    with mock.patch.object(rt_flomni, "feedback_is_running", return_value=False):
        with mock.patch.object(rt_flomni, "laser_tracker_on") as laser_tracker_on:
            with pytest.raises(RtError):
                rt_flomni.feedback_enable_without_reset()
                laser_tracker_on.assert_called_once()
                assert mock.call(b"l3\n") in rt_flomni.sock.put.mock_calls
